from guppy.heapy.test import support

PORTABLE_TEST = 1       # Relax tests to be more portable


class TestCase(support.TestCase):
    def setUp(self):
        support.TestCase.setUp(self)
        self.View.is_rg_update_all = False

        self.US = US = self.heapy.UniSet

        self.Use = Use = self.heapy.Use
        Use.reprefix = 'hp.'
        self.do = lambda x: x.dictof

        self.un = Use.Anything.fam
        self.ty = Use.Type
        self.rc = Use.Rcs
        self.iso = Use.iso

        self.Anything = US.Anything
        self.Nothing = US.Nothing

        class C1:
            def x(self):
                return 0

        class C2:
            pass
        c1 = C1()

        self.C1 = C1
        self.C2 = C2
        self.c1 = c1

    def lt(self, a, b):
        self.assertTrue(a < b)

    def eq(self, a, b):
        self.assertTrue(a == b)

    def dj(self, a, b):
        # disjoint; not related by <= or >=, and not overlapping
        self.assertTrue(not a <= b)
        self.assertTrue(not b <= a)
        self.assertTrue(not a & b)
        self.assertTrue(a.disjoint(b))

    def nr(self, a, b):
        # not related by <= or >=, and overlapping
        self.assertTrue(not a <= b)
        self.assertTrue(not b <= a)
        self.assertTrue(a & b)
        self.assertTrue(not a.disjoint(b))


class SpecialCases(TestCase):
    # Special tests that catch cases that came up during development & debugging
    def test_1(self):
        un = self.un
        ty = self.ty
        do = self.do
        rc = self.rc
        iso = self.iso
        All = self.Anything
        Nothing = self.Nothing
        C1 = self.C1
        C2 = self.C2
        c1 = self.c1

        def eq(a, b):
            self.assertTrue(a == b)
            self.assertTrue(str(a) == str(b))

        e1 = []
        e2 = {}
        e3 = []
        e4 = ()

        a = ty(int)
        b = ty(dict)
        self.assertTrue(~a & ~b != Nothing)

        eq(ty(list) & iso(e1, e2, e3), iso(e1, e3))
        eq((ty(list) | ty(dict)) & iso(e1, e2, e3, e4), iso(e1, e2, e3))
        eq(iso(e1, e3) | ty(list), ty(list))
        eq(ty(list) | iso(e1, e3), ty(list))

        eq(iso(e1, e3) - iso(e3), iso(e1))
        eq(~iso(e3) & iso(e1, e3), iso(e1))

        eq(iso(e1, e2, e3) - ty(dict), iso(e1, e3))
        eq(~ty(dict) & iso(e1, e2, e3), iso(e1, e3))
        eq(ty(dict) | iso(e1, e2), ty(dict) | iso(e1))
        eq(iso(e1, e2) | ty(dict), ty(dict) | iso(e1))
        eq((ty(dict) | ty(tuple)) | iso(e1, e2), (ty(dict) | ty(tuple)) | iso(e1))
        eq(iso(e1, e2) | (ty(dict) | ty(tuple)), (ty(dict) | ty(tuple)) | iso(e1))
        eq(~ty(dict) | iso(e1, e2), ~ty(dict) | iso(e2))
        eq(iso(e1, e2) | ~ty(dict), ~ty(dict) | iso(e2))
        eq(ty(dict) - iso(e1, e2), ty(dict) - iso(e2))
        eq(~iso(e1, e2) & ty(dict), ty(dict) - iso(e2))

        eq(iso(e1, e3) ^ iso(e2), iso(e1, e2, e3))
        eq(iso(e1, e3) ^ iso(e2, e3), iso(e1, e2))
        eq(iso(e1, e3) ^ iso(e1, e3), Nothing)

        eq(iso(e1, e3) <= ty(list), True)
        eq(iso(e1, e2) <= ty(list) | ty(dict), True)
        eq(ty(list) >= iso(e1, e3), True)
        eq(ty(list) | ty(dict) >= iso(e1, e2), True)

    def test_2(self):
        un = self.un
        ty = self.ty
        do = self.do
        rc = self.rc
        iso = self.iso
        All = self.Anything
        Nothing = self.Nothing
        C1 = self.C1
        C2 = self.C2
        c1 = self.c1

        class C3(object):
            def x(self):
                return 1

        def asrt(x):
            self.assertTrue(x)

        def no(x):
            self.assertTrue(not x)

        eq = self.aseq

        # Tests to do with Nothing being finite - having length and iteration

        no(dict in (ty(dict) | ty(int)))
        no([] in (ty(dict) | ty(int)))
        asrt({} in (ty(dict) | ty(int)))
        asrt(dict in (ty(dict) | ty(int) | ty(type(dict))))
        asrt(list(ty(list) & iso({})) == [])

        # When creating ISO classes, we don't want to memoize them
        # which would leak the elements.

        from sys import getrefcount as grc
        import sys
        import types

        c = C1()
        rc = grc(c)
        x = iso(c)
        x = None
        eq(grc(c), rc)

    def test_dictowner(self):
        # Special test for dict ownership
        # motivated by: dicts that are not found in traversal, should not
        # cause repeated (unsuccessfull) updates of dict ownership
        # This is a performance question, requires special kind of testing
        #
        # Also tests that dict & dict owners are not leaked

        import gc
        from sys import getrefcount as grc
        Use = self.Use
        C1 = self.C1
        c1 = self.c1
        iso = self.iso

        o = self.python.io.StringIO()

        # Create a dict hidden from view
        d1 = self.View.immnodeset([{}])
        d3 = {}

        # Remember the initial ref counts for target objects

        gc.collect()

        rcd1 = grc(list(d1)[0])
        rcd3 = grc(d3)
        rcC1 = grc(C1)
        rcc1 = grc(c1)
        rcdc1 = grc(c1.__dict__)

        clock = self.python.time.process_time

        N = 5
        M = 50

        # This was the fast case, when only reachable dicts are classified
        for i in range(N):
            print(iso(d3).kind, file=o)
            print(iso(c1.__dict__).kind, file=o)

        # Now measure it

        while 1:
            gc.collect()
            t = clock()
            for i in range(M):
                iso(d3).kind
                iso(c1.__dict__).kind
            fast = clock()-t
            if fast >= 0.5:  # Enough resolution?
                break
            else:
                M *= 2  # No, try more loops

        # This was a slow case; involving repeated classification of a unreachable dict
        # It was originally 4.97 times slower when N was 5
        # The problem occurs for successive classifications of different dicts,
        # when at least one of them is unreachable.

        gc.collect()
        for i in range(N):
            print(iso(*d1).kind, file=o)
            print(iso(c1.__dict__).kind, file=o)

        gc.collect()
        # Now measure it

        t = clock()
        for i in range(M):
            iso(*d1).kind
            iso(c1.__dict__).kind
        slow = clock()-t

        self.assertTrue(slow <= 1.5*fast)

        # This is another slow case according to notes Nov 18 2004.
        # A succession of different unreachable dicts.

        gc.collect()
        dn = self.View.immnodeset([{} for i in range(N)])
        for i in range(N):
            print(iso(list(dn)[i]).kind, file=o)

        # Now measure it
        gc.collect()
        dn = self.View.immnodeset([{} for i in range(M)])

        t = clock()
        for i in range(M):
            iso(list(dn)[i]).kind
        slow = clock()-t

        # Sometimes M might be huge, and the vast majority of time it is not
        # doing the classification.
        # self.assertTrue(slow <= 1.5*fast)

        # Partition was likewise slow for unreachable dicts
        dn = self.View.immnodeset([{} for i in range(N)])
        gc.collect()
        print([x[0] for x in Use.Clodo.classifier.partition(dn)], file=o)

        # Now measure it
        dn = self.View.immnodeset([{} for i in range(M)])
        gc.collect()
        t = clock()
        [x[0] for x in Use.Clodo.classifier.partition(dn)]
        slow = clock()-t
        self.assertTrue(slow <= 1.5*fast)

        # Check that ref counts for target objects are the same as initially

        gc.collect()
        gc.collect()    # Note May 17 2005

        self.aseq(grc(list(d1)[0]), rcd1)
        self.aseq(grc(d3), rcd3)
        self.aseq(grc(c1), rcc1)
        self.aseq(grc(C1), rcC1)
        self.aseq(grc(c1.__dict__), rcdc1)

        self.aseq(o.getvalue(), """\
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict of <Module>.C1
dict (no owner)
dict (no owner)
dict (no owner)
dict (no owner)
dict (no owner)
[hp.Nothing.dictof]
""".replace('<Module>', self.__module__))

    def test_retclaset(self):
        # Test (A) that referrer classifications don't leak their classes
        # and (B) that selection is not disturbed by list arguments
        # (This is removed since it doesnt always work)
        # and (C) that selection does update referrer graph correctly

        self.__module__ = '<Module>'  # Make the rendering independent on our name

        from sys import getrefcount as grc
        import gc
        C1 = self.C1
        c1 = self.c1

        iso = self.iso
        rcC1 = grc(C1)

        o = self.python.io.StringIO()
        print(iso(C1).byrcs.kind, file=o)

        s = iso(c1).byrcs.kind
        print(s, file=o)
        self.aseq(s & iso(c1), iso(c1))

        x = C1()

        self.aseq(s & iso(c1, x), iso(c1))

        s = iso(x).byrcs.kind
        self.aseq(s & iso(c1, x), iso(x))
        x = C1()
        # (C) make sure referrer graph is updated by select
        self.aseq(s & iso(c1, x), iso(x))

        s = None
        x = None
        gc.collect()
        gc.collect()                    # Note May 17 2005
        self.aseq(grc(C1), rcC1)        # (A)

    def test_alt_retclaset(self):
        # Test the alternative referrer memo update
        # On low level, and the speed of selection

        import gc
        iso = self.iso
        a = []
        b = self.View.immnodeset([[]])

        x = [a, b]

        hv = self.View.hv

        rg = self.View.nodegraph()

        gc.collect()
        hv.update_referrers_completely(rg)
        self.assertTrue(x in rg[a])

        self.assertTrue(rg[list(b)[0]] == (None,))
        rg.clear()
        rg = None

        # Test View functionality

        self.View.is_rg_update_all = True
        gc.collect()
        iso(a).referrers
        self.assertTrue(a in self.View.rg.get_domain())
        self.assertTrue(list(b)[0] in self.View.rg.get_domain())

        clock = self.python.time.process_time
        s = iso(a)
        N = 1000
        while 1:
            t = clock()
            for i in range(N):
                s.referrers
            fast = clock()-t
            if fast >= 0.5:
                break
            N *= 2      # CPU is too fast to get good resolution, try more loops

        t = clock()
        for i in range(N):
            self.View.rg.domain_covers([a])
            self.View.rg[a]
        faster = clock()-t
        s = iso(*b)
        t = clock()
        for i in range(N):
            s.referrers
        slow = clock() - t
        self.assertTrue(not slow > fast * 4)

    def test_via(self, vlist=['v', ]):  # vlist is just to make v unoptimizable
        # Special tests for the via classifier

        from sys import getrefcount as grc
        import gc

        iso = self.iso
        hp = self.Use
        d = {}
        k = ('k',)
        v = tuple(vlist)  # Make sure v is not optimized to a constant

        d[k] = v
        d[v] = v

        rck = grc(k)
        rcv = grc(v)

        s = iso(v)

        self.assertTrue(s.byvia.kind == hp.Via("_.f_locals['v']", "_[('k',)]", "_[('v',)]", '_.keys()[1]') or
                        s.byvia.kind == hp.Via("_.f_locals['v']", "_[('k',)]", "_[('v',)]", '_.keys()[0]'))

        del s
        gc.collect()
        gc.collect()
        self.aseq(grc(k), rck)
        self.aseq(grc(v), rcv)


class RenderCase(TestCase):
    def test_rendering(self):
        import sys
        import types
        iso = self.iso
        C1 = self.C1
        c1 = self.c1

        class C3(object):
            def x(self):
                return 1

        e1 = []
        e2 = {}
        e3 = []

        o = self.python.io.StringIO()
        # str'ing of homogenous & inhoumogenous values

        self.US.summary_str.str_address = lambda x: '<address>'

        def ps(x):
            print(x.brief, file=o)

        ps(iso(1, 2))
        ps(iso(1, 2.0, 3.0))
        ps(iso(e1))
        ps(iso(e1, e2))
        ps(iso(e1, e3))

        ps(iso(self.python.builtins.TypeError()))
        # ps(iso(type('MetaType', (type,), {})('MetaTypeIns', (), {})))
        ps(iso(None))
        ps(iso(sys, support, types))
        ps(iso(int, type, C3))
        ps(iso(C1()))
        ps(iso(C3()))
        ps(iso(C1))
        ps(iso(C3))
        ps(iso(len))
        ps(iso(self.setUp))
        ps(iso(C1.x))
        ps(iso(C1().x))
        ps(iso(C3.x))
        ps(iso(C3().x))

        ps(iso({}))
        ps(iso(c1.__dict__))
        ps(iso(types.__dict__))

        try:
            1/0
        except ZeroDivisionError:
            typ, value, traceback = sys.exc_info()

        ps(iso(traceback))
        ps(iso(traceback.tb_frame))

        expected = """\
<2 int: 1, 2>
<3 (float | int): <2 float: 2.0, 3.0> | <1 int: 1>>
<1 list: <address>*0>
<2 (dict (no owner) | list): <1 dict (no owner): <address>*0> | <1 list: <ad...>
<2 list: <address>*0, <address>*0>
<1 TypeError: <address>>
<1 builtins.NoneType: None>
<3 module: guppy.heapy.test.support, sys, types>
<3 type: <Module>.C3, int, type>
<1 <Module>.C1: <address>>
<1 <Module>.C3: <address>>
<1 type: <Module>.C1>
<1 type: <Module>.C3>
<1 types.BuiltinMethodType: len>
<1 types.MethodType: <<Module>.RenderCase at <addre...>
<1 function: <Module>.x>
<1 types.MethodType: <<Module>.C1 at <address>>.x>
<1 function: <Module>.x>
<1 types.MethodType: <<Module>.C3 at <address>>.x>
<1 dict (no owner): <address>*0>
<1 dict of <Module>.C1: <address>>
<1 dict of module: types>
<1 types.TracebackType: <in frame <test_rendering at <address>> at <address>>>
<1 types.FrameType: <test_rendering at <address>>>
""".replace('<Module>', self.__module__)
        self.aseq(o.getvalue(), expected)

        if PORTABLE_TEST:
            return

        o = self.python.io.StringIO()

        # The following is nonportable, sizes may change
        # In particular, the list size changed from 2.3 to 2.4
        # The following test is only for 2.3 in 32-bit python

        # pp'ing prints in a nice form
        # This tests all types currently defined in Classifiers.Summary_str
        # and then some
        # Except: frametype; its size varies from time to time!

        x = iso(len, C1, 1.0+3j, {1: 2, 3: 4}, 1.25, C1.x.__func__, 1, ['list'],
                100000000000, None, C1.x, C1().x, C3.x, C3().x, sys, support,
                'string', ('tuple',), C3, int, type(None),
                # and some types not defined
                C1(), C3(), c1.__dict__

                )

        print(x, file=o)
        print(x.more, file=o)

        # Test instancetype; we need to replace the classifier with bytype

        x = iso(C1()).bytype
        print(x, file=o)

        expected = """\
Partition of a set of 24 objects. Total size = 2128 bytes.
 Index  Count   %     Size   % Cumulative  % Kind (class / dict of class)
     0      3  12     1272  60      1272  60 type
     1      4  17      144   7      1416  67 types.MethodType
     2      1   4      136   6      1552  73 dict (no owner)
     3      1   4      136   6      1688  79 dict of <Module>.C1
     4      1   4       60   3      1748  82 list
     5      1   4       56   3      1804  85 function
     6      2   8       48   2      1852  87 module
     7      1   4       44   2      1896  89 class
     8      1   4       32   2      1928  91 <Module>.C1
     9      1   4       32   2      1960  92 str
<8 more rows. Type e.g. '_.more' to view.>
 Index  Count   %     Size   % Cumulative  % Kind (class / dict of class)
    10      1   4       32   2      1992  94 types.BuiltinMethodType
    11      1   4       28   1      2020  95 <Module>.C3
    12      1   4       28   1      2048  96 tuple
    13      1   4       24   1      2072  97 complex
    14      1   4       20   1      2092  98 long
    15      1   4       16   1      2108  99 float
    16      1   4       12   1      2120 100 int
    17      1   4        8   0      2128 100 types.NoneType
Partition of a set of 1 object. Total size = 32 bytes.
 Index  Count   %     Size   % Cumulative  % Type
     0      1 100       32 100        32 100 types.InstanceType
""".replace('<Module>', self.__module__)
        self.aseq(o.getvalue(), expected)


class BaseCase(TestCase):
    def test_minmax(self):
        s = self.guppy.sets.immbitset
        min = self.US.minimals
        max = self.US.maximals

        self.aseq(min([]), [])
        self.aseq(min([1]), [1])
        self.aseq(min([1, 1]), [1])
        self.aseq(min([1, 2]), [1])
        self.aseq(min([[], []]), [[]])

        self.aseq(min([s([1]), s([1, 2])]), [s([1])])
        self.aseq(min([s([1]), s([1, 2]), s([3])]), [s([1]), s([3])])

        self.aseq(max([]), [])
        self.aseq(max([1]), [1])
        self.aseq(max([1, 1]), [1])
        self.aseq(max([1, 2]), [2])
        self.aseq(max([[], []]), [[]])

        self.aseq(max([s([1]), s([1, 2])]), [s([1, 2])])
        self.aseq(max([s([1]), s([1, 2]), s([3])]), [s([1, 2]), s([3])])

    def test_base_classes(self):
        un = self.un
        ty = self.ty
        do = self.do
        rc = self.rc
        iso = self.iso
        All = self.Anything
        Nothing = self.Nothing
        C1 = self.C1
        C2 = self.C2
        c1 = self.c1
        lt = self.lt
        eq = self.eq
        dj = self.dj
        nr = self.nr

        data = [
            (All,       eq,     All),


            (ty(int),   eq,     ty(int)),
            (ty(int),   dj,     ty(dict)),
            (ty(int),   lt,     All),

            (rc(ty(dict)), eq,   rc(ty(dict))),
            (rc(ty(dict)), lt,   All),
            (rc(ty(dict)), dj,   rc(ty(list))),

            (iso(1),    eq,     iso(1)),
            (iso(1),    lt,     All),
            (iso(1),    dj,     iso(2)),
            (iso(1),    lt,     ty(int)),
            (iso(1),    dj,     ty(dict)),

            (Nothing,   eq,     Nothing),
            (Nothing,   lt,     ty(int)),
            (Nothing,   lt,     iso(1)),
        ]

        # Test relation of base classifications
        for a, cmp, b in data:
            cmp(a, b)
            # Test the four set-operations: & | - ^
            # depending on the asserted argument relation
            if cmp is eq:
                eq(b, a)
            elif cmp is lt:
                self.assertTrue(b > a)
                eq(b ^ a, b - a)        # Simple transformation
                eq(a ^ b, b - a)        # -=-, indep. of type
                lt(a, b)
            elif cmp is dj:
                dj(b, a)  # check that the dj relation is symmetric
                eq(a & b, Nothing)
                eq(b & a, Nothing)
                eq(a | b, b | a)
                eq(a - b, a)
                eq((a | b) - b, a)
                eq(a ^ b, a | b)
                eq(b ^ a, a | b)
                lt(a, a | b)
                lt(b, a | b)
            elif cmp is nr:
                nr(b, a)         # symmetric as well
                eq(a & b, b & a)
                eq(a & b & b, a & b)
                eq((a & b) - b, Nothing)
                eq((a | b) - b, a - b)
                eq(a | b, b | a)
                lt(Nothing, a & b)
                lt(Nothing, b & a)
                lt(a & b, a)
                lt(a & b, b)
                lt(a - b, a)
                dj(a - b, b)
                lt(a ^ b, a | b)
                lt(a, a | b)
                lt(b, a | b)

    def test_invalid_operations(self):
        US = self.US
        US.auto_convert_iter = False
        US.auto_convert_type = False

        ty = self.ty
        c1 = self.c1

        self.assertRaises(TypeError, lambda: ty(c1))
        self.assertRaises(TypeError, lambda: ty(int) <= None)
        self.assertRaises(TypeError, lambda: None >= ty(int))
        self.assertRaises(TypeError, lambda: None <= ty(int))

        self.assertRaises(TypeError, lambda: list(ty(int)))
        self.assertRaises(TypeError, lambda: len(ty(int)))

        self.assertRaises(TypeError, lambda: ty(int) & None)
        self.assertRaises(TypeError, lambda: None & ty(int))
        self.assertRaises(TypeError, lambda: ty(int) | None)
        self.assertRaises(TypeError, lambda: None | ty(int))
        self.assertRaises(TypeError, lambda: ty(int) - None)
        self.assertRaises(TypeError, lambda: None - ty(int))
        self.assertRaises(TypeError, lambda: ty(int) ^ None)
        self.assertRaises(TypeError, lambda: None ^ ty(int))

        self.assertRaises(TypeError, lambda: ty(int) | [14])
        self.assertRaises(TypeError, lambda: ty(int) | dict)
        self.assertRaises(TypeError, lambda: ty(int) | self.C1)

    def test_fancy_list_args(self):
        # Test the, normally disabled, possibility to use iterables as
        # right and left arguments in set expressions.
        # This option can cause problems as noted 22/11 2004.

        self.US.auto_convert_iter = True

        eq = self.eq
        iso = self.iso
        ty = self.ty

        e1 = []
        e2 = {}
        e3 = []
        e4 = ()

        eq(ty(list) & [e1, e2, e3], iso(e1, e3))
        eq([e1, e2, e3] & ty(list), iso(e1, e3))       # Requires __rand__
        eq([e1, e2, e4] & (ty(dict) | ty(list)) == [e1, e2], True)
        eq([e1, e2] & (ty(dict) | ty(list)) == [e1, e2], True)
        eq(iso(e1, e2) & (ty(dict) | ty(list)) == [e1, e2], True)
        eq(iso(e1, e2) & [e1, e3], iso(e1))
        eq(iso(e1, e2) | [e1, e3], iso(e1, e2, e3))
        # Requires __ror__
        eq([e1, e3] | iso(e1, e2), iso(e1, e2, e3))
        eq(iso(e1, e3) - [e3], iso(e1))
        eq([e1, e3] - iso(e3), iso(e1))                 # Requires __rsub__
        eq([e1, e2, e3] - ty(dict), iso(e1, e3))
        eq(~ty(dict) & [e1, e2, e3], iso(e1, e3))
        eq(iso(e1, e3) ^ [e2], iso(e1, e2, e3))
        eq([e2] ^ iso(e1, e3), iso(e1, e2, e3))           # Requires __rxor__
        eq([e1, e2] <= iso(e1, e2, e3), True)
        eq([e1, e2] <= ty(list) | ty(dict), True)
        eq((ty(list) | ty(dict)) >= [e1, e2], True)
        eq([e1, e2] <= ty(list), False)
        eq([e1, e2] <= iso(e1), False)
        eq([e1, e2] >= iso(e1, e2, e3), False)
        eq([e1, e2] >= iso(e1, e2), True)
        eq(iso(e1, e2, e3) <= [e1, e2], False)
        eq(iso(e1, e2) <= [e1, e2], True)
        eq(iso(e1, e2, e3) >= [e1, e2], True)
        eq(iso(e1, e2) >= [e1, e2, e3], False)

    def test_fancy_type_conversions(self):
        # Test the, perhaps optional, possibility to use types and classes
        # in classification set expressions.

        self.US.auto_convert_type = True

        un = self.un
        ty = self.ty
        do = self.do
        rc = self.rc
        iso = self.iso
        All = self.Anything
        Nothing = self.Nothing
        C1 = self.C1
        C2 = self.C2
        c1 = self.c1

        def eq(a, b):
            self.assertTrue(a == b)

        e1 = []
        e2 = {}
        e3 = []
        e4 = ()

        eq(ty(dict), dict)
        eq(iso(e1, e2) & dict, iso(e2))
        eq(dict & iso(e1, e2), iso(e2))
        eq(iso(e1, e2) | dict, iso(e1) | ty(dict))
        eq(dict | iso(e1, e2), iso(e1) | ty(dict))
        eq(iso(e1, e2) - dict, iso(e1))
        eq(dict - iso(e1, e2), ty(dict) - iso(e2))
        eq(iso(e1, e2, e3) ^ dict, (ty(dict)-iso(e2)) | iso(e1, e3))


class LawsCase(TestCase):
    def test_laws(self):
        un = self.un
        ty = self.ty
        do = self.do
        rc = self.rc
        iso = self.iso
        All = self.Anything
        Nothing = self.Nothing
        C1 = self.C1
        C2 = self.C2
        c1 = self.c1
        lt = self.lt
        eq = self.eq

        t = self.guppy.sets.test

        absorption = t.absorption
        associative = t.associative
        commutative = t.commutative
        deMorgan = t.deMorgan
        distributive = t.distributive
        idempotence = t.idempotence
        inclusion = t.inclusion

        def ltr(a, b, level=3):
            lt(a, b)
            eq(a & b, a)
            eq(b & a, a)
            eq(a | b, b)
            eq(b | a, b)
            eq(a - b, Nothing)
            eqr(b - a, b - a)
            eq((b - a) | a, b)
            eq(a | (b - a), b)
            eq(a & (b - a), Nothing)
            eq((b - a) & a, Nothing)
            eq((b - a) - a, (b - a))
            eq(a - (b - a), a)  # note Nov 3 2004
            if level > 0:
                if a is Nothing:
                    eq(b - a, b)
                else:
                    ltr(b - a, b, level-1)

        def eqr(a, b, level=1):
            eq(a, b)
            eq(a & b, a)
            eq(a | b, a)
            eq(a - b, Nothing)
            eq(a ^ b, Nothing)
            if level:
                eqr(b, a, level - 1)

        classes = [All, ty(int), ty(type(c1)), rc(ty(dict)), iso(c1), Nothing]

        for a in classes:
            idempotence(a)
            for b in classes:
                if a <= b:
                    if b <= a:
                        eqr(a, b)
                    else:
                        ltr(a, b)
                elif b <= a:
                    ltr(b, a)

                absorption(a, b)
                commutative(a, b)

                inclusion(a, b)
                deMorgan(a, b)
                for c in classes:
                    associative(a, b, c)
                    deMorgan(a, b, c)
                    distributive(a, b, c)


class ClassificationCase(TestCase):
    def test_classification(self):
        # Test classification by the standard classifiers
        self.View.is_rg_update_all = True  # Tricky details Note Apr 22 2005
        Use = self.Use
        iso = self.iso
        nodeset = self.heapy.UniSet.immnodeset

        class A:
            pass

        class B(object):
            pass
        a = A()
        b = B()
        li = [1, [], {}, a, b, a.__dict__, b.__dict__]
        for o in li:
            self.asis(iso(o).bytype.kind.arg, type(o))
        for o in li:
            if o is a.__dict__:
                kind = iso(a).kind
            elif o is b.__dict__:
                kind = iso(b).kind
            elif type(o) is dict:
                kind = Use.Nothing
            elif o is a:
                kind = a.__class__
            else:
                kind = type(o)
            self.aseq(iso(o).kind.arg, kind)
        cla = iso(()).byunity.kind
        self.asis(cla.arg, None)
        for o in li:
            self.aseq(iso(o).byunity.kind, cla)
        for o in li:
            self.aseq(iso(o).byid.kind, Use.Id(id(o)))
        # self.View.update_referrers(nodeset(li))
        for i, o in enumerate(li):
            cl = iso(o).byrcs.kind
            if 1 <= i <= 2:
                self.aseq(cl, Use.Clodo.sokind(list).refdby)
            if i == 5:
                self.aseq(cl, Use.Clodo.sokind(A)(list).refdby)
            if i == 6:
                self.aseq(cl, Use.Clodo.sokind(B)(list).refdby)

    def test_selection(self):
        # Test classifications operations via selection invariant

        Use = self.Use

        class A:
            pass

        class B(object):
            pass
        a = A()
        b = B()
        li = Use.iso(135, [], {}, a, b, a.__dict__, b.__dict__)

        allers = (Use.Unity, Use.Type, Use.Clodo,
                  Use.Rcs, Use.Via)  # , Use.Id
        ps = {}
        for er in allers:
            # p = er.classifier.partition(li.nodes)
            p = [(av.kind, av) for av in li.by(er).partition]
            for ak, av in p:
                if ak in ps:
                    self.aseq(ps[ak],  av)
                else:
                    ps[ak] = av

        for ak, av in list(ps.items()):
            self.aseq(ak & li, av)
            for bk, bv in list(ps.items()):
                # Test set operations by selection definition
                self.aseq((ak & bk) & li, av & bv)
                self.aseq((ak | bk) & li, av | bv)
                self.aseq((ak - bk) & li, av - bv)
                self.aseq((bk - ak) & li, bv - av)
                self.aseq((ak ^ bk) & li, av ^ bv)


def test_main(testrender=1, debug=0):
    support.run_unittest(BaseCase, debug)
    support.run_unittest(ClassificationCase, debug)
    support.run_unittest(LawsCase, debug)
    support.run_unittest(RenderCase, debug)
    support.run_unittest(SpecialCases, debug)



""" Support functions for RE simplification.
This module is intended for use by the RE module.
It is in a separate module to keep RE itself cleaner
since the algorithm, 'rectangle selection', is a quite
separate part that depends on some tricky heuristics.

The primary entry function is

chooserects(lines, gauges)

It chooses 'the best' rectangles from lines to base simplification on.
A weight on atoms is given by gauges.

pr() gives some example usages of chooserects.

"""

from guppy.sets import immbitset, mutbitset, immbitrange


class Rect(object):
    __slots__ = 'width', 'lines', 'gainmemo', 'lnos', 'all_lines', 'common_part'

    def __init__(self, width, lines):
        self.width = width
        self.lines = lines
        assert not (width and len(lines) == 1)
        self.gainmemo = {}

    def init2(self, lnobyid, all_lines):
        self.all_lines = all_lines
        self.lnos = immbitset([lnobyid[id(line)] for line in self.lines])
        self.common_part = self.get_common_part()

    def reducelines(self, lnos):
        # Reduce lines of self by removing some lines
        # Argument: lnos, a 'set' of line numbers to remove

        olnos = self.lnos
        lnos = olnos & ~lnos
        if lnos != olnos:
            self.lnos = lnos
            self.lines = [self.all_lines[lno] for lno in lnos]
        if len(lnos) == 1:
            self.width = len(self.lines[0])

    def get_lines(self, pickednos=0):
        lines = []
        for i in self.lnos & ~ pickednos:
            lines.append(self.all_lines[i])
        return lines

    def __repr__(self):
        return '<\n dir = %d\n width = %d\n lnos = %s\n lines = %s\n>' % (
            self.dir, self.width, list(self.lnos), self.lines)


class LeftRect(Rect):
    __slots__ = ()
    dir = 0

    def get_common_part(self):
        return self.lines[0][:self.width]

    def get_uncommons(self, pickednos=0):
        uc = []
        for line in self.get_lines(pickednos):
            uc.append(line[self.width:])
        return uc


class RightRect(Rect, object):
    __slots__ = ()
    dir = -1

    def get_common_part(self):
        lo = -self.width
        if lo == 0:
            return []
        return self.lines[0][lo:]

    def get_uncommons(self, pickednos=0):
        uc = []
        hi = -self.width
        if hi == 0:
            hi = None
        for line in self.get_lines(pickednos):
            uc.append(line[:hi])
        return uc


def sum_gauge(gauge, lst):
    global hits, misses
    if gauge is None:
        return len(lst)
    else:
        gain = 0
        for x in lst:
            gain += gauge(x)
    return gain


def cmp_gauged(xs, ys, gauges):
    for gauge in gauges:
        gx = sum_gauge(gauge, xs)
        gy = sum_gauge(gauge, ys)
        c = cmp(gx, gy)
        if c:
            return c
    return 0


class InducedRect:
    def __init__(self, s, lines, lnos):
        self.s = s
        self.width = s.width
        self.all_lines = s.all_lines
        self.lines = lines
        self.lnos = lnos


class InducedRightRect(InducedRect, RightRect):
    pass


class InducedLeftRect(InducedRect, LeftRect):
    pass


def brect(lines):
    if len(lines) <= 1:
        return [LeftRect(0, lines)]
    newrects = [LeftRect(0, lines), RightRect(0, lines)]
    donerects = []
    while newrects:
        oldrects = newrects
        newrects = []
        for r in oldrects:
            width = r.width
            while 1:
                is_done = 0
                d = {}
                pos = width ^ r.dir
                for line in r.lines:
                    if width < len(line):
                        d.setdefault(line[pos], []).append(line)
                    else:
                        is_done = 1
                if is_done or len(d) != 1:
                    break
                r.width = width = width + 1
            donerects.append(r)
            width += 1
            for k, v in list(d.items()):
                if len(v) > 1:
                    new_r = r.__class__(width, v)
                    newrects.append(new_r)

    return donerects


def choose(rects, lines=[], gauges=[None]):
    def induce(r):
        uncommons = r.get_uncommons()
        if len(uncommons) < 2:
            return
        irs = []
        for s in rects:
            if s.dir != r.dir:
                continue
            pss = []
            uncs = s.get_uncommons(pickednos)
            lnos = s.lnos & ~pickednos
            assert len(uncs) == len(lnos)
            for unc, lno in zip(uncs, lnos):
                if unc in uncommons:
                    pss.append(lno)
            if len(pss) == len(uncommons):
                pslnos = immbitset(pss)
                pss = [lines[lno] for lno in pss]
                if s.dir == -1:
                    c = InducedRightRect
                else:
                    c = InducedLeftRect
                ir = c(s, pss, pslnos)
                irs.append(ir)

        if irs:
            news.extend(irs)

    def overlap(r):
        rlnos = r.lnos
        tonews = []
        for s in rects:
            if s is r:
                continue
            if s.dir != r.dir:
                continue
            slnos = s.lnos
            if not (slnos & rlnos):
                continue
            slnos &= ~ pickednos
            if not slnos:
                # remove
                continue
            scom = s.common_part
            if not scom:
                continue
            for t in rects:
                if t is s:
                    continue
                if t.dir == s.dir:
                    continue
                tlnos = t.lnos & ~pickednos
                if (tlnos & rlnos):
                    continue
                olnos = tlnos & slnos
                if not olnos:
                    continue
                if slnos == tlnos:
                    continue
                tcom = t.common_part
                if not tcom:
                    continue
                c = cmp_gauged(scom, tcom, gauges)
                if c > 0:
                    continue

                break

            else:
                # s is ok
                tonews.append(s)
                rects.remove(s)
        news.extend(tonews)

    def picknext():
        while 1:
            if news:
                r = news[0]
                del news[0]
            else:
                r = None
                for s in list(rects):
                    slnos = s.lnos - pickednos
                    if not slnos:
                        rects.remove(s)
                        continue
                    sn = len(slnos) - 1
                    sw = s.width
                    if r is not None:
                        if not sw:
                            break
                        if not sn:
                            continue
                        if rwn:
                            rmemo = r.gainmemo
                            smemo = s.gainmemo
                            c = 0
                            for gauge in gauges:
                                try:
                                    gr = rmemo[gauge]
                                except KeyError:
                                    gr = sum_gauge(gauge, r.common_part)
                                    rmemo[gauge] = gr
                                gr *= rn
                                try:
                                    gs = smemo[gauge]
                                except KeyError:
                                    gs = sum_gauge(gauge, s.common_part)
                                    smemo[gauge] = gs
                                gs *= sn
                                c = gr - gs
                                if c:
                                    break
                            if c >= 0:
                                continue
                    r = s
                    rlnos = slnos
                    if not sw:
                        break
                    rn = sn
                    rw = sw
                    rwn = sn * sw
                if r is not None:
                    rects.remove(r)
            if r is not None:
                r.reducelines(pickednos)
                if r.lnos:
                    return r

    def cmpinit(x):
        return x.width, x.dir, x.lnos[0]

    if gauges[0] is None:
        gauges = gauges[1:]

    lnobyid = dict([(id(line), i) for i, line in enumerate(lines)])

    orects = rects
    rects = list(orects)

    for r in rects:
        r.init2(lnobyid, lines)

    rects.sort(key=cmpinit, reverse=True)

    allnos = immbitrange(len(lines))

    pickednos = mutbitset()
    pickedrects = []

    news = []
    while pickednos != allnos:
        r = picknext()
        pickednos |= r.lnos
        pickedrects.append(r)
        induce(r)
        overlap(r)

    return pickedrects


def chooserects(lines, gauges=[None]):
    rects = brect(lines)
    choosen = choose(rects, lines, gauges)
    return choosen


def pr():
    x = chooserects(['abc', 'ade'])
    x = chooserects(['abc',
                     'abe',
                     'ace',
                     'xby'])

    print(x)
    x = chooserects(['ab1',
                     'ab2',
                     'ac3',
                     'ac4'])

    print(x)

    # Case where.. (from right).. :
    # the total gain of two rects (bfbf+cfcf = 4) is greater than the gain of another
    # overlapping rect (ffff == 3), although the individual gains are less (= 2).
    # But in this case at least, the end result should likely become the same

    x = chooserects([
        '1bf',
        '2bf',
        '3cf',
        '4cf'])

    print(x)

    # Case where it chooses twice..

    x = chooserects([
        'abc',
        'abd',
        'bcx',
        'bdy'
    ])

    print('TW', x)

    # Case where it didn't choose enough rects

    x = ([
        'abc',
        'abd',
        'e'
    ])

    print(chooserects(x))

    # Case where it should prefer one side or the other
    # i.e. left traditionally

    print(chooserects(['abc', 'axc']))

    # Case where it should give a width 0 rect

    print(chooserects(['a', '']))

    # Case with overlap

    print(chooserects(['abcd', 'abce', 'a', 'f']))

    # Case with induce

    print(chooserects(['abcd', 'abce', 'd', 'e']))

    print(chooserects(['auvw', 'buvw', 'a', 'b']))
    print(chooserects(['axuvw', 'bxuvw', 'axy', 'bxy', 'cy']))


def tmany():
    for i in range(100):
        x = chooserects(['abc',
                         'abe',
                         'ace',
                         'xby'])



# Convert a node representation to text
# in some different forms

# o output to an object with an interface subset of Tk Text
# o output to ascii-only - best attempt w.o. different fonts
# o output to man-page text

# Parameterized on out

# I have to define the roles of responsibility.

# out is a thin layer above something Tk text-like
# it isolates some details but doesn't provide real new functionality

# Node2Inter takes care of the general formatting
# It needs to know about
# o the size of out, i.e. width and height
# o the fonts and sizes available
# o the special characters available
# o the size of the fonts
# o we restrict to fixed size fonts
#   otherwise it would have to ask for size of strings?
#

# to be able to do
# o line wrapping
# o paragraphs
# o headers
# o item and bullet lists
# o tables

# The out can 'record' its commands
# and play them on an actual Tk text window.
# with a simple translation
# o it can operate stack-free
# -- so after eg a font change, there is a font change back

# Does the out has a configuration table
# such as,
# o the fonts to choose
# o for different sizes
# o whether to make items bold

# In any case there may be default here..


class Node2Inter:
    def __init__(self, mod, node, out, cnf, width=None):
        self.mod = mod
        self.out = out
        self.cnf = cnf
        self.width = width

        self.span_stack = []
        self.attrs_stack = [{}]
        self.atparbegin = 1
        self.inpre = 0
        self.ul_level = 0
        self.prev_margin_bottom = 0

        self.sizescale = cnf.sizescale

        if node is not None:
            self._visit_outer_node(node)

    def _visit_outer_node(self, node):
        cnf = self.cnf
        self.span_begin(font_family=cnf.textfamily, font_size=cnf.sizeindex)
        node, attrs = node.split_attrs()
        for k, v in attrs:
            k = k.strip()
            v = v.strip()
            setattr(self.out, '_gsl_%s' % k, v)
        node.accept(self)
        self.span_end()

    def _visit_node(self):
        node, attrs = node.split_attrs()
        if attrs:
            self.attrs_stack

    def append(self, x):
        self.out.insert('end', x, self.tags)

    def div(self, node, margin_bottom=0, **kwds):
        self.div_begin(**kwds)
        node.arg_accept(self)
        self.div_end(margin_bottom)

    def div_begin(self, margin_top=0, **kwds):
        if self.span_stack:
            d = self.span_stack[-1]
            if 'spacing1' in d:
                if self.atparbegin:
                    margin_top = max(margin_top, d['spacing1'])
                self.tag_config(self.tag, **d)
                del d['spacing1']

        margin_top = max(margin_top, self.prev_margin_bottom)
        self.prev_margin_bottom = 0

        if not self.atparbegin:
            self.nl()
            self.atparbegin = 1

        if margin_top:
            kwds['spacing1'] = margin_top

        self.span_begin(**kwds)

    def div_end(self, margin_bottom=0):
        if not self.atparbegin:
            self.nl()
        self.span_end()
        self.atparbegin = 1
        self.prev_margin_bottom = margin_bottom

    def getopt(self, name, default=0):
        if self.span_stack and name in self.span_stack[-1]:
            return self.span_stack[-1][name]
        else:
            return getattr(self, name, default)

    def nl(self):
        self.append('\n')

    def set_default_tag(self):
        if self.span_stack:
            tag = 't%s' % self.mod._root.pickle.dumps(self.span_stack[-1])
        else:
            tag = 'tag'
        self.tag = tag
        self.tags = (tag,)

    def span(self, node, **kwds):
        self.span_begin(**kwds)
        node.arg_accept(self)
        self.span_end()

    def span_begin(self, **kwds):
        if self.span_stack:
            d = self.span_stack[-1].copy()
            d.update(kwds)
        else:
            d = kwds
        self.span_stack.append(d)

        self.set_default_tag()

    def span_end(self):
        tag = self.tag
        self.tag_config(tag, **self.span_stack[-1])
        self.span_stack.pop()
        self.set_default_tag()

    def tag_config(self, tag, **kwds):
        okwds = {}
        fontspecs = []
        for k, v in list(kwds.items()):
            if not k.startswith('font_'):
                okwds[k] = v
                continue
            else:
                fontspecs.append((k[5:], v))
        if fontspecs:
            font = [None, None, '']
            for k, v in fontspecs:
                if k == 'size':
                    v = max(0, min(len(self.sizescale)-1, v))
                    font[1] = self.sizescale[v]
                elif k == 'family':
                    font[0] = v
                else:
                    if font[2]:
                        font[2] += ' '
                    font[2] += k
            if not font[2]:
                font.pop()
            okwds['font'] = tuple(font)

        self.out.tag_config(tag, **okwds)

    def text(self, text):
        if not self.inpre:
            if self.atparbegin:
                text = text.lstrip()
            if not text:
                return

            text = text.replace('\n', ' ')
            text = text.replace('\t', ' ')
            while '  ' in text:
                text = text.replace('  ', ' ')

            if self.atparbegin and self.prev_margin_bottom:
                self.tag_config(self.tag, **self.span_stack[-1])
                self.span_stack[-1]['spacing1'] = self.prev_margin_bottom
                self.set_default_tag()
                self.prev_margin_bottom = 0

            self.append(text)
        else:
            text = text.expandtabs()
            idx = text.find('\n')
            if idx != -1 and 'spacing1' in self.span_stack[-1]:
                self.append(text[:idx+1])
                self.tag_config(self.tag, **self.span_stack[-1])
                del self.span_stack[-1]['spacing1']
                self.set_default_tag()
                text = text[idx+1:]
            if text:
                self.append(text)

        self.atparbegin = 0

    def _visit_children(self, node):
        E = self.mod.ReportedError
        for ch in node.children:
            try:
                ch.accept(self)
            except E:
                pass

    def _visit_hx(self, node):
        n = int(node.tag[1:])
        font_size = 7 - n
        margin_top = 12 - 1 * n
        margin_bottom = 12 - 1 * n

        self.div(node,
                 font_size=font_size,
                 font_bold=1,
                 margin_top=margin_top,
                 margin_bottom=margin_bottom)

    def visit_big(self, node):
        self.span(node, font_size=self.getopt('font_size') + 1)

    def visit_blockquote(self, node):
        lmargin = self.getopt('lmargin1') + 36
        rmargin = self.getopt('rmargin') + 36
        self.div(node,
                 lmargin1=lmargin,
                 lmargin2=lmargin,
                 rmargin=rmargin,
                 margin_top=6,
                 margin_bottom=6,
                 )

    def visit_char(self, node):
        code = node.arg.strip()
        if code == 'nbsp':
            self.span_begin(invisible=1)
            self.append('x')
            self.span_end()
        else:
            self.error(
                'I do not know how to render this character code: %r.' % code, node)

    def visit_code(self, node):
        self.span(node, font_family=self.cnf.codefamily)

    def visit_comment(self, node):
        pass

    def visit_dl(self, node):
        self.div(node)

    def visit_dt(self, node):
        self.div(node)

    def visit_dd(self, node):
        lmargin = self.getopt('lmargin1') + 36
        self.div(node, lmargin1=lmargin, lmargin2=lmargin)

    def visit_define(self, node):
        # xxx
        self._visit_children(node)

    def visit_div(self, node):
        self.div(node)

    def visit_document(self, node):
        self._visit_children(node)

    def visit_document_lang(self, node):
        if self.document_lang is not None:
            self.error('Duplicate document lang directive.', node)
        self.document_lang = node

    def visit_document_title(self, node):
        if self.document_title is not None:
            self.error('Duplicate document title directive.', node)
        self.document_title = node

    def visit_em(self, node):
        self.span(node, font_italic=1)

    def visit_file(self, node):
        self._visit_children(node)

    def visit_gsl_title(self, node):
        self.out._gsl_title = node.arg

    def visit_gsl_width(self, node):
        self.out._gsl_width = int(node.arg)

    def visit_gsl_height(self, node):
        self.out._gsl_height = int(node.arg)

    def visit_h1(self, node):
        self._visit_hx(node)

    def visit_h2(self, node):
        self._visit_hx(node)

    def visit_h3(self, node):
        self._visit_hx(node)

    def visit_h4(self, node):
        self._visit_hx(node)

    def visit_h5(self, node):
        self._visit_hx(node)

    def visit_h6(self, node):
        self._visit_hx(node)

    def visit_li(self, node):
        indent = self.getopt('lmargin1') + 18
        self.div_begin(
            lmargin1=indent,
            lmargin2=indent
        )

        mode = ['disc', 'square', 'circle'][self.ul_level % 3]

        char = {'disc': '*', 'circle': 'O', 'square': '[]'}[mode]

        self.span_begin()
        self.text('%s ' % char)
        self.span_end()
        self.span_begin(
            lmargin1=indent,
            lmargin2=indent+12
        )
        node.arg_accept(self)
        self.span_end()
        self.div_end()

    def visit_p(self, node):
        self.div(node, margin_top=6, margin_bottom=6)

    def visit_pre(self, node):
        self.inpre += 1
        self.div(node, font_family=self.cnf.codefamily,
                 margin_top=6, margin_bottom=6)
        self.inpre -= 1

    def visit_small(self, node):
        self.span(node, font_size=self.getopt('font_size') - 1)

    def visit_span(self, node):
        self.span(node)

    def visit_string(self, node):
        self._visit_children(node)

    def visit_strong(self, node):
        self.span(node, font_bold=1)

    def visit_sub(self, node):
        self.span(node,
                  font_size=self.getopt('font_size') - 1,
                  offset=self.getopt('offset') - 2
                  )

    def visit_sup(self, node):
        self.span(node,
                  font_size=self.getopt('font_size') - 1,
                  offset=self.getopt('offset') + 2
                  )

    def visit_table(self, node):
        Table(self, node)
        pass

    def visit_td(self, node):
        pass

    def visit_th(self, node):
        pass

    def visit_tr(self, node):
        pass

    def visit_text(self, node):
        self.text(node.arg)
        self._visit_children(node)

    def visit_u(self, node):
        self.span(node, underline=1)

    def visit_ul(self, node):
        self.ul_level += 1
        self.div(node)
        self.ul_level -= 1

    def visit_var(self, node):
        self.span(node, font_italic=1)


class SimulText:
    def __init__(self, mod, width=None):
        self.mod = mod
        self.width = width

        self.lines = [[]]
        self.tags = {}
        self.textntags = []
        self.fonts = {}

    def insert(self, pos, text, tags):
        assert pos == 'end'
        lines = text.split('\n')
        self.lines[-1].append((lines[0], tags))
        for line in lines[1:]:
            self.lines.append([(line, tags)])
        self.textntags.append((text, tags))

    def tag_config(self, tag, **kwds):
        if tag in self.tags and kwds == self.tags[tag]:
            return
        self.tags[tag] = kwds

    ##

    def finalize(self):
        if len(self.lines[-1]) == 1 and not self.lines[-1][0][0]:
            self.lines.pop()
        if self.width is not None:
            self.wrap_lines()

    def get_width(self):
        width = 0
        for line in self.lines:
            w = self.text_width(line)
            if w > width:
                width = w
        return width

    def replay(self, out, lineidx):
        if lineidx >= len(self.lines):
            return
        line = self.lines[lineidx]
        for (ch, tags) in line:
            out.insert('end', ch, tags)
            for tag in tags:
                out.tag_config(tag, **self.tags[tag])

    def split_word(self, line):
        words = [[]]
        for text, tags in line:
            wtext = text.split(' ')
            for wt in wtext:
                if wt:
                    words[-1].append((wt, tags))
                if words[-1]:
                    words.append([])
        return words

    def text_width(self, textntags):
        font = None
        subline = None
        subfonts = []
        for ch, tags in textntags:
            for tag in tags:
                if tag in self.tags and 'font' in self.tags[tag]:
                    newfont = self.tags[tag]['font']
                    break
            else:
                assert 0

            if newfont != font:
                if subline:
                    subfonts.append((subline, font))
                font = newfont
                subline = []
            subline.append(ch)
        if subline:
            subfonts.append((subline, font))
        width = 0
        for (subline, font) in subfonts:
            f = self.mod.makefont(font)
            m = f.measure(''.join(subline))
            width += m
        return width

    def width_to(self, char):
        # distance from left margin to first occurence of char
        # or the width of longest line, if char not found
        for line in self.lines:
            w = 0
            found = 0
            for (text, tags) in line:
                if char in text:
                    text = text[:text.index(char)]
                    found = 1
                w += self.text_width([(text, tags)])
                if found:
                    break
            if found:
                break
        if not found:
            w = self.get_width()
        return w

    def wrap_line(self, line):
        w = self.text_width(line)
        if w <= self.width:
            self.lines.append(line)
            return

        words = self.split_word(line)

        i = 0
        while i < len(words):
            pre = list(words[i])
            w = self.text_width(pre)
            while w > self.width:
                # The word is too long to fit.
                # I have to cut it off.
                # xxx this may be somewhat slow
                #  measuring after every character
                j = 0
                # Position j at the chunk that is going to be split
                while j + 1 < len(pre):
                    w = self.text_width(pre[:j+1])
                    if w > self.width:
                        break
                    j += 1

                # Split this chunk
                # Allow at least one character
                k = 2
                while k <= len(pre[j][0]):
                    w = self.text_width(
                        pre[:j-1] + [(pre[j][0][:k], pre[j][1])])
                    if w > self.width:
                        break
                    k += 1
                self.lines.append(pre[:j-1] + [(pre[j][0][:k-1], pre[j][1])])
                assert self.text_width(self.lines[-1]) <= self.width
                pre = [(pre[j][0][k-1:], pre[j][1])]
                w = self.text_width(pre)

            i += 1
            while i < len(words):
                space = [(' ', pre[-1][1])]
                word = words[i]
                w = self.text_width(pre + space + word)
                if w > self.width:
                    break
                else:
                    pre.extend(space + word)
                    i += 1
            self.lines.append(pre)

    def wrap_lines(self):
        lines = self.lines
        self.lines = []
        for line in lines:
            self.wrap_line(line)


class TableCell:
    def __init__(self, row, node):
        self.row = row
        self.table = row.table
        self.parent = self.table.parent
        self.cnf = self.parent.cnf
        self.mod = self.parent.mod

        self.attrs = {}
        self.node = self.set_attributes(node)

        self.gen_out()

    def align(self, pos, width):
        align = self.attrs['align']
        if align == 'center':
            self.tabstop = (pos + 0.5*width, 'center')
        elif align == 'left':
            self.tabstop = (pos, 'left')
        elif align == 'right':
            self.tabstop = (pos+width, 'right')
        elif align == 'char':
            w = self.out.width_to(self.attrs['char'])
            co = float(self.attrs['charoff'].rstrip('%'))/100.0
            self.tabstop = (pos + co*width-w, 'left')
        elif align == 'justify':
            # XXX I don't know how this works
            self.tabstop = (pos + 0.5*width, 'center')
        else:
            raise ValueError('Invalid align: %s' % align)

    def get_edges(self, width):
        align = self.attrs['align']
        mywidth = self.width
        if align == 'center':
            l, r = 0.5 * width - 0.5 * mywidth, 0.5 * width + 0.5 * mywidth
        elif align == 'left':
            l, r = 0,  mywidth
        elif align == 'right':
            l, r = width - mywidth, width
        elif align == 'char':
            w = self.out.width_to(self.attrs['char'])
            co = float(self.attrs['charoff'].rstrip('%'))/100.0
            l = co * width - w
            r = l + mywidth
        elif align == 'justify':
            # XXX I don't know how this works
            l, r = 0, width
        else:
            raise ValueError('Invalid align: %s' % align)
        return l, r

    def get_width(self):
        self.width = self.out.get_width()
        self.numlines = len(self.out.lines)
        return self.width

    def set_attributes(self, node):
        a = self.attrs
        if node.tag == 'th':
            align = 'center'
        else:
            align = 'left'
        a['align'] = align
        a['char'] = self.cnf.decimal_point
        a['charoff'] = '50%'

        node, attrs = node.split_attrs()
        for k, v in attrs:
            a[k] = v

        return node

    def gen_out(self, width=None):

        self.out = SimulText(self.mod, width=width)

        n2i = Node2Inter(self.mod, None, self.out, self.cnf, width=width)

        kwds = self.parent.span_stack[-1].copy()
        if self.node.tag == 'th':
            kwds['font_bold'] = 1

        n2i.span_begin(**kwds)
        self.node.arg_accept(n2i)
        n2i.span_end()
        self.out.finalize()

        self.get_width()

    def wrap_to_width(self, width):
        if width >= self.width:
            return
        self.gen_out(width)


class TableRow:
    def __init__(self, table, node):
        self.table = table
        self.node = node
        self.numlines = 1

        self.cells = []

        node, attrs = node.split_attrs()
        self.attrs = attrs
        node.children_accept(self)

    def new_cell(self, node):
        cell = TableCell(self, node)
        self.cells.append(cell)

    def visit_td(self, node):
        self.new_cell(node)

    def visit_th(self, node):
        self.new_cell(node)


class Table:
    def __init__(self, parent, node):
        self.parent = parent
        self.node = node

        self.caption = None
        self.rows = []

        parent.div_begin(margin_top=6)

        self.lmargin = parent.getopt('lmargin1')

        node.children_accept(self)

        Width = 400
        w = self.columnify()
        widths = self.widths
        spacings = self.spacings

        if w > Width:
            # Which one to wrap?
            # The longest?
            # All?
            gw = [Width / len(self.widths)]*len(self.widths)
            extra = 0
            others = list(range(len(self.widths)))
            for i, w in enumerate(self.widths):
                if w < gw[i]:
                    extra += gw[i] - w
                    gw[i] = w
                    others.remove(i)
            extra = int(extra / len(others))
            for i in others:
                gw[i] += extra

            widths = self.widths = gw
            for row in self.rows:
                col = 0
                for cell in row.cells:
                    cell.wrap_to_width(gw[col])
                    col += 1

            for row in self.rows:
                col = 0
                pos = 0
                for cell in row.cells:
                    w = widths[col]
                    cell.align(pos+self.lmargin, w)
                    pos += w + spacings[col]
                    col += 1
                    row.numlines = max(row.numlines, cell.numlines)

        for row in self.rows:
            for i in range(row.numlines):
                tabstops = []
                for cell in row.cells:
                    tabstops.extend(cell.tabstop)
                tabstops = tuple(tabstops)
                if i == 0 and row is self.rows[0]:
                    tabkwds = row.cells[0].out.tags[row.cells[0].out.lines[0][0][1][0]]
                else:
                    tabkwds = {}
                    if row is not self.rows[0] and i == 0:
                        tabkwds['spacing1'] = 6
                tabtag = str(tabstops)+str(tabkwds)
                for cell in row.cells:
                    parent.out.insert('end', '\t', (tabtag,))
                    cell.out.replay(parent.out, i)
                parent.out.tag_config(tabtag, tabs=tabstops, **tabkwds)
                parent.nl()
        parent.div_end()

    def columnify(self):
        # Make the cells aligned in columns

        widths = self.widths = []

        for row in self.rows:
            col = 0
            for cell in row.cells:
                w = cell.get_width()
                if col >= len(widths):
                    widths.append(w)
                else:
                    widths[col] = max(w, widths[col])
                row.numlines = max(row.numlines, cell.numlines)

                col += 1

        # Extra spacing after column i
        spacings = self.spacings = [0] * len(widths)

        MINSPACING = 10

        for row in self.rows:
            col = 0
            for cell in row.cells[:-1]:
                rcell = row.cells[col+1]
                ledge = cell.get_edges(widths[col])[1]
                redge = rcell.get_edges(widths[col+1])[0]+widths[col]
                spacing = MINSPACING - (redge - ledge)
                spacings[col] = max(spacing, spacings[col])
                col += 1

        width = 0
        for row in self.rows:
            col = 0
            pos = 0
            for cell in row.cells:
                w = widths[col]
                cell.align(pos+self.lmargin, w)
                pos += w + spacings[col]
                col += 1

            if pos > width:
                width = pos

        self.width = width
        return width

    def visit_tfoot(self, node):
        node.children_accept(self)

    def visit_thead(self, node):
        node.children_accept(self)

    def visit_tr(self, node):
        row = TableRow(self, node)
        self.rows.append(row)


class RecordingInter:
    FLATTEXT = 1
    FLATKWDS = 0
    lasttext = ()
    lasttag = None
    def __init__(self):

        self.appends = []
        self.tag_configs = {}
        self.lasttext = []
        self.clearmemo()

    def __str__(self):
        return 'APPENDS: %s TAG_CONFIGS: %s' % (self.appends, self.tag_configs)

    def clearmemo(self):
        self.memo = {}          # Maps any value to it self
        self.tagmemo = {}       # Maps tag to integer tag number

    def flush(self):
        if self.lasttext:
            tag = self.tagmemo.setdefault(self.lasttag, len(self.tagmemo))
            text = ''.join(self.lasttext)
            text = self.memo.setdefault(text, text)
            if self.FLATTEXT:
                self.appends.append(tag)
                self.appends.append(text)
            else:
                tt = tag, text
                tt = self.memo.setdefault(tt, tt)
                self.appends.append(tt)
            self.lasttext = []

    def insert(self, pos, text, tags):
        assert pos == 'end'
        assert len(tags) == 1

        tag = tags[0]
        if tag != self.lasttag:
            self.flush()
            self.lasttag = tag
        self.lasttext.append(text)

    def play(self, out):
        self.flush()
        if self.FLATTEXT:
            i = 0
            while i < len(self.appends):
                tag = self.appends[i]
                text = self.appends[i+1]
                out.insert('end', text, (tag,))
                i += 2
        else:
            for tag, text in self.appends:
                out.insert('end', text, (tag,))
        for (tag, kwdlist) in list(self.tag_configs.items()):

            if self.FLATKWDS:
                kwds = {}
                i = 0
                while i < len(kwdlist):
                    kwds[kwdlist[i]] = kwdlist[i+1]
                    i += 2
                out.tag_config(tag, **kwds)
            else:
                out.tag_config(tag, **dict(kwdlist))

        for k in self.__dict__:
            if k.startswith('_gsl_'):
                setattr(out, k, getattr(self, k))

    def prepare_for_pickle(self):
        # Call this before pickling to reduce space usage.
        self.flush()
        for k in list(self.__dict__.keys()):
            if k not in ('appends', 'tag_configs') and not k.startswith('_gsl_'):
                delattr(self, k)

    def tag_config(self, tag, **kwds):
        kwdlist = []
        for k, v in list(kwds.items()):
            k = self.memo.setdefault(k, k)
            v = self.memo.setdefault(v, v)
            if self.FLATKWDS:
                kwdlist.append(k)
                kwdlist.append(v)
            else:
                kv = k, v
                kv = self.memo.setdefault(kv, kv)
                kwdlist.append(kv)
        kwdlist = tuple(kwdlist)
        kwdlist = self.memo.setdefault(kwdlist, kwdlist)
        tag = self.tagmemo.setdefault(tag, len(self.tagmemo))

        if tag in self.tag_configs:
            assert self.tag_configs[tag] == kwdlist
        else:
            self.tag_configs[tag] = kwdlist


class TextInter:

    def __init__(self, mod, wid):
        self.mod = mod
        self.wid = wid

        for name in (
            'config',
            'insert',
            'tag_delete',
        ):
            setattr(self, name, getattr(wid, name))

    def tag_config(self, tag, **kwds):
        if 'invisible' in kwds:
            del kwds['invisible']
            kwds['foreground'] = kwds['background'] = kwds.get(
                'background', self.wid['background'])

        self.wid.tag_config(tag, **kwds)


class TkConfig:
    sizeindex = 3
    sizescale = (6, 8, 10, 12, 16, 20, 24, 28)
    textfamily = 'times'
    codefamily = 'courier'
    decimal_point = '.'  # default CHAR attribute


class _GLUECLAMP_:
    _imports_ = (
        '_parent:SpecNodes',
        '_parent.SpecNodes:node_of_taci',
        '_parent.SpecNodes:node_of_string',
        '_parent.Main:ReportedError',
        '_parent:Html',
        '_root:pickle',
        '_root.md5:md5',
        '_root:os',
        '_root:re',
        '_root:string',
        '_root:tkinter',
    )

    def _get_makefont(self):
        fonts = {}
        root = self.tkinter.Tk()
        root.withdraw()

        def makefont(font):

            if font in fonts:
                return fonts[font]
            weight = 'normal'
            slant = 'roman'
            if len(font) > 2:
                if 'bold' in font[2]:
                    weight = 'bold'
                if 'italic' in font[2]:
                    slant = 'italic'
            f = self.tkinter.font.Font(family=font[0], size=font[1],
                                 weight=weight, slant=slant)
            fonts[font] = f
            return f
        return makefont

    def _get_tkconfig(self):
        return TkConfig()

    def node2inter(self, node, inter, tkconfig=None):
        if tkconfig is None:
            tkconfig = self.tkconfig
        Node2Inter(self, node, inter, tkconfig)

    def gsltextviewer(self, parent=None, filename=None, text=None, node=None, htmloutfile=None,
                      inpickle=0,
                      inrecorder=0,
                      outrecorder=0
                      ):

        # It seems they dont want we mix data and py files in the dist sigh
        # so these are last minute hacks

        pickle = self.pickle

        if inpickle:
            inrecorder = pickle.loads(inpickle)

        if node is None:
            if text is None:
                if filename is not None:
                    with open(filename) as f:
                        text = f.read()
            node = self.node_of_string(text, nostrip=1)

        if htmloutfile is not None:
            self.Html.node2file(node, htmloutfile)

        if outrecorder:
            r = RecordingInter()
            self.node2inter(node, r)
            r.prepare_for_pickle()
            return r

        cache = None
        if filename is not None:
            sp = self.os.path.splitext(filename)
            if sp[1] == '.gsl':
                cache = sp[0] + '.gsc'

        m = self._root.guppy.etc.textView.TextViewer(
            parent, 'Untitled', data='')
        v = m.textView
        v['state'] = 'normal'
        v['font'] = 'Times -12'
        v.bind('<Destroy>', lambda event: m.quit())

        if cache or inrecorder:
            if inrecorder:
                r = inrecorder
            else:
                r = None
                textdigest = self.md5(text.encode('utf-8')).digest()
                try:
                    f = open(cache)
                except IOError:
                    pass
                else:
                    td = f.read(len(textdigest))
                    if td == textdigest:
                        r = pickle.load(f)
                    f.close()
                if r is None:
                    r = RecordingInter()
                    self.node2inter(node, r)
                    r.prepare_for_pickle()
                    f = open(cache, 'w')
                    try:
                        try:
                            f.write(textdigest)
                        except IOError:
                            pass  # maybe write protected just ignore for now XXX
                        else:
                            pickle.dump(r, f, 0)
                    finally:
                        f.close()
            r.play(v)

        else:
            self.node2inter(node, v)

        title = getattr(v, '_gsl_title', None)
        if title:
            m.title(title)
            m.iconname(title)

        geometry = getattr(v, '_gsl_tk_geometry', None)
        if geometry:
            m.geometry(geometry)

        v['state'] = 'disabled'
        return m


def test_string(s=None, name=None):
    from guppy import Root
    gsl = Root().guppy.gsl
    me = gsl.Text
    if s is None:
        s = getattr(me._parent.test.testdata, name)

    T = me.tkinter
    node = me.node_of_string(s, nostrip=1)

    me._parent.Html.node2file(node, '/tmp/x.html')

    t = RecordingInter()
    me.node2inter(node, t)

    t.prepare_for_pickle()

    root = T.Tk()
    root.withdraw()
    text = me._root.guppy.etc.textView.TextViewer(
        root, 'test', data='').textView
    text['state'] = 'normal'
    text['font'] = 'Times -12'
    text.bind('<Destroy>', lambda event: root.quit())
    ti = TextInter(me, text)

    t.play(ti)

    text.mainloop()


def test():
    name = 'long_wrapping_tables'
    name = 'html_tables'
    test_string(name=name)



# -*- coding: utf-8 -*-

import io
import inspect
import os

THISDIR = os.path.dirname(__file__)


class GuppyDoc:
    def __init__(self, mod, str):
        self.mod = mod
        if str is None:
            str = '???'

        str = self._substitute(str)

        self.str = str

    def __getitem__(self, idx):
        refs = {}
        refline = False
        for line in self.str.split('\n'):
            if line == 'References':
                refline = True
                continue
            elif not refline:
                continue
            line = line.strip()
            assert line.startswith('[')
            refstop = line.index(']')
            ref = line[1:refstop]
            refs[int(ref)] = line[refstop+1:].strip()
        url = refs[idx]
        url = self.mod.docurl(url)
        return self.mod.open_browser(url)

    def getheader(self):
        lines = self.str.split('\n')
        header = []
        for line in lines:
            if not line:
                break
            header.append(line)
        return '\n'.join(header)

    def _substitute(self, str):
        rp = self.mod._root.guppy.heapy.Use.reprefix
        str = str.replace('$HP.', rp)
        return str

    def __repr__(self):
        return self.str


class Lister:
    def __init__(self):
        self.output = io.StringIO()

    def list(self, items, columns=4, width=80):
        items = items[:]
        colw = width // columns
        rows = (len(items) + columns - 1) // columns
        for row in range(rows):
            for col in range(columns):
                i = col * rows + row
                if i < len(items):
                    self.output.write(items[i])
                    if col < columns - 1:
                        self.output.write(' ' + ' ' * (colw-1 - len(items[i])))
            self.output.write('\n')
        return self

    def getvalue(self):
        return self.output.getvalue()


class GuppyDir(object):
    def __init__(self, li, obj, mod, opts='', **kwds):
        self.li = li
        self.obj = obj
        self.mod = mod
        self.opts = opts
        self.kwds = kwds

    def __call__(self, opts=None):
        li = self.li
        obj = self.obj
        mod = self.mod
        if opts is None:
            opts = self.opts
        return self.__class__(li, obj, mod, opts)

    def __getattr__(self, attr):
        return self.mod.getdoc2(self.obj, attr)

    def __getitem__(self, idx):
        return self.li[idx]

    def __repr__(self):
        opts = self.opts
        if 'L' in opts:
            r = ''
            for d in self.li:
                r += '*** ' + d + ' ***\n' + repr(getattr(self, d))+'\n\n'
        elif 'l' in opts:
            r = ''
            for d in self.li:
                t = getattr(self, d).getheader()
                if not (t.startswith(d) or t.startswith('x.'+d)):
                    t = d
                r += t + '\n\n'
        else:
            r = Lister().list(self.li).getvalue().rstrip()
        header = self.kwds.get('header', '')
        if header:
            header += '\n'
        footer = self.kwds.get('footer', '')
        if footer:
            footer = '\n'+footer
        r = header+r+footer
        return r


class _GLUECLAMP_:
    def dir(self, obj=None, opts='', **kwds):
        try:
            share = obj._share
        except AttributeError:
            return self.getdir_no_share(obj, opts, **kwds)
        clamp = share.Clamp
        dl = getattr(clamp, '_dir_', None)
        if dl is not None:
            dl = list(dl)
        else:
            dl = []
            private = getattr(clamp, '_private_', ())
            try:
                imports = clamp._imports_
            except AttributeError:
                pass
            for imp in imports:
                ix = imp.find(':')
                if ix == -1:
                    continue
                dl.append(imp[ix+1:])
            for gm in dir(clamp):
                if gm.startswith('_get_'):
                    dl.append(gm[5:])
                else:
                    if not gm.startswith('_'):
                        dl.append(gm)
            dl = [d for d in dl if not d in private]
        dl.sort()
        return GuppyDir(dl, obj, self, opts, **kwds)

    def getdir_no_share(self, obj, opts, **kwds):
        dl = dir(obj)
        dl = [d for d in dl if not d.startswith('_')]
        return GuppyDir(dl, obj, self, opts, **kwds)

    def getdoc2(self, obj, name):
        try:
            share = obj._share
        except AttributeError:
            return self.getdoc_no_share(obj, name)
        clamp = obj._share.Clamp
        try:
            imports = clamp._imports_
        except AttributeError:
            pass
        else:
            for imp in imports:
                ix = imp.find(':')
                if ix == -1:
                    pass
                else:
                    if imp[ix+1:] == name:
                        return self.getdoc_import(obj, clamp, name, imp, ix)
        for gm in dir(clamp):
            if gm.startswith('_get_') and gm[5:] == name:
                return self.getdoc__get_(clamp, gm)
            else:
                if name == gm:
                    return self.getdoc_other(clamp, name)

        return GuppyDoc(self, '???')

    def getdoc_no_share(self, obj, name):
        try:
            doc = getattr(obj, '_doc_'+name)
        except AttributeError:
            pass
        else:
            return GuppyDoc(self, doc)

        cl = obj.__class__
        p = getattr(cl, name)
        if isinstance(p, property):
            docobj = p
        else:
            docobj = getattr(obj, name)

        return self.getdoc_obj(docobj)

    def getdoc__get_(self, clamp, gm):
        func = getattr(clamp, gm)
        doc = func.__doc__
        return GuppyDoc(self, doc)

    def getdoc_import(self, obj, clamp, name, imp, ix):
        doc = ''
        if hasattr(clamp, '_doc_'+name):
            doc = getattr(obj, '_doc_'+name)
        else:
            impobj = getattr(obj, imp[ix+1:])
            doc = getattr(impobj, '__doc__')
        return GuppyDoc(self, doc)

    def getdoc_obj(self, obj):
        doc = inspect.getdoc(obj)
        if doc is None:
            doc = '???'
        return GuppyDoc(self, doc)

    def getdoc_other(self, obj, name):
        attr = getattr(obj, name)
        doc = inspect.getdoc(attr)
        if doc:
            return GuppyDoc(self, doc)

        try:
            doc = getattr(obj, '_doc_'+name)
        except AttributeError:
            doc = ''
        if doc is None:
            doc = '?'
        print('doc', doc)
        return GuppyDoc(self, doc)

    def docurl(self, url):
        url = 'https://zhuyifei1999.github.io/guppy3/'+url
        return url

    def open_browser(self, url):
        import webbrowser
        webbrowser.open(url)
